package cbase

import (
	"fmt"
	"os"
	"os/signal"
	"sync"
	"syscall"

	"go_dreamfactory/lego"
	"go_dreamfactory/lego/core"
	"go_dreamfactory/lego/sys/event"
	"go_dreamfactory/lego/sys/log"
)

//模块的运行载体对象 管理模块的配置以及协程对象
type defaultModule struct {
	seetring map[string]interface{}
	mi       core.IModule
	closeSig chan bool
	wg       sync.WaitGroup
}

//模块的独立协程
func (this *defaultModule) run() {
	this.wg.Add(1)
	this.mi.Run(this.closeSig)
	this.wg.Done()
}

//模块的销毁逻辑
func (this *defaultModule) destroy() (err error) {
	defer lego.Recover(fmt.Sprintf("Module :%s destroy", this.mi.GetType()))
	err = this.mi.Destroy()
	if err != nil {
		err = fmt.Errorf("关闭模块【%s】失败 err:%s", this.mi.GetType(), err.Error())
	}
	this.wg.Wait()
	return
}

///基础服务对象
type ServiceBase struct {
	closesig chan string                        //服务关闭信号管道
	service  core.IService                      //服务的上层实现对象
	comps    map[core.S_Comps]core.IServiceComp //组件列表
	modules  map[core.M_Modules]*defaultModule  //模块列表
}

//服务初始化过程 完成 系统 和 服务组件的 初始化过程
func (this *ServiceBase) Init(service core.IService) (err error) {
	this.closesig = make(chan string, 1)
	this.service = service
	this.modules = make(map[core.M_Modules]*defaultModule)
	this.service.InitSys()
	for _, v := range this.comps {
		options := v.NewOptions()
		if o, ok := service.GetSettings().Comps[string(v.GetName())]; ok {
			options.LoadConfig(o)
		}
		err = v.Init(this.service, v, options)
		if err != nil {
			return
		}
	}
	log.Infof("服务[%s] 初始化完成!", this.service.GetId())
	return nil
}

//配置服务组件
func (this *ServiceBase) OnInstallComp(cops ...core.IServiceComp) {
	this.comps = make(map[core.S_Comps]core.IServiceComp)
	for _, v := range cops {
		if _, ok := this.comps[v.GetName()]; ok {
			log.Errorf("覆盖注册组件【%s】", v.GetName())
		}
		this.comps[v.GetName()] = v
	}
}

//启动服务 完成服务组件的启动过程
func (this *ServiceBase) Start() (err error) {
	for _, v := range this.comps {
		err = v.Start()
		if err != nil {
			return
		}
	}
	log.Infof("服务[%s:%s] 启动完成!", this.service.GetId(), this.service.GetVersion())
	return
}

//运行模块
func (this *ServiceBase) Run(mod ...core.IModule) {
	for _, v := range mod { //初始化模块对象 分配配置参数
		if sf, ok := this.service.GetSettings().Modules[string(v.GetType())]; ok {
			this.modules[v.GetType()] = &defaultModule{
				seetring: sf,
				mi:       v,
				closeSig: make(chan bool, 1),
			}
		} else {
			this.modules[v.GetType()] = &defaultModule{
				seetring: make(map[string]interface{}),
				mi:       v,
				closeSig: make(chan bool, 1),
			}
			log.Warnf("注册模块【%s】 没有对应的配置信息", v.GetType())
		}
	}
	for _, v := range this.modules { //序列化每一个模块的参数对象 完成模块的初始化 过程
		options := v.mi.NewOptions()
		if err := options.LoadConfig(v.seetring); err == nil {
			err := v.mi.Init(this.service, v.mi, options)
			if err != nil {
				log.Panicf(fmt.Sprintf("初始化模块【%s】错误 err:%v", v.mi.GetType(), err))
			}
		} else {
			log.Panicf(fmt.Sprintf("模块【%s】 Options:%v 配置错误 err:%v", v.mi.GetType(), v.seetring, err))
		}
	}
	for _, v := range this.modules { //完成模块的启动过程
		err := v.mi.Start()
		if err != nil {
			log.Panicf(fmt.Sprintf("启动模块【%s】错误 err:%v", v.mi.GetType(), err))
		}
	}
	for _, v := range this.modules { //启动模块的独立协程
		go v.run()
	}
	event.TriggerEvent(core.Event_ServiceStartEnd) //广播服务启动完毕事件
	//监听外部关闭服务信号
	c := make(chan os.Signal, 1)
	//添加进程结束信号
	signal.Notify(c,
		os.Interrupt,    //退出信号 ctrl+c退出
		os.Kill,         //kill 信号
		syscall.SIGHUP,  //终端控制进程结束(终端连接断开)
		syscall.SIGINT,  //用户发送INTR字符(Ctrl+C)触发
		syscall.SIGTERM, //结束程序(可以被捕获、阻塞或忽略)
		syscall.SIGQUIT) //用户发送QUIT字符(Ctrl+/)触发
	select {
	case sig := <-c:
		log.Errorf("服务[%s] 关闭 signal = %v\n", this.service.GetId(), sig)
	case <-this.closesig:
		log.Errorf("服务[%s] 关闭\n", this.service.GetId())
	}
}

//完毕服务模块接口
func (this *ServiceBase) Close(closemsg string) {
	this.closesig <- closemsg
}

//服务关闭时的消息调用过程
func (this *ServiceBase) Destroy() (err error) {
	for _, v := range this.modules {
		v.closeSig <- true
		v.wg.Wait()
		err = v.destroy()
		if err != nil {
			return
		}
	}
	for _, v := range this.comps {
		err = v.Destroy()
		if err != nil {
			return
		}
	}
	return
}

//获取当前服务下指定模块对象
func (this *ServiceBase) GetModule(ModuleName core.M_Modules) (module core.IModule, err error) {
	if v, ok := this.modules[ModuleName]; ok {
		return v.mi, nil
	} else {
		return nil, fmt.Errorf("未装配模块【%s】", ModuleName)
	}
}

//获取当前服务下指定组件对象
func (this *ServiceBase) GetComp(CompName core.S_Comps) (comp core.IServiceComp, err error) {
	if v, ok := this.comps[CompName]; ok {
		return v, nil
	} else {
		return nil, fmt.Errorf("Service 未装配组件【%s】", CompName)
	}
}
