package core

type S_Category string  //服务类别 例如 网关服务 游戏服务 业务服务   主要用于服务功能分类
type M_Modules string   //模块类型
type S_Comps string     //服务器组件类型
type ErrorCode int32    //错误码
type Event_Key string   //事件Key
type Rpc_Key string     //RPC
type Redis_Key string   //Redis缓存
type SqlTable string    //数据库表定义
type CustomRoute uint8  //自定义网关

const (
	AutoIp = "0.0.0.0"         //自动ip 可以匹配任意ip地址
	AllIp  = "255.255.255.255" //多选ip 匹配全部ip
)

const ( //默认事件
	Event_ServiceStartEnd  Event_Key = "ServiceStartEnd"  //服务完全启动完毕
	Event_FindNewService   Event_Key = "FindNewService"   //发现新的服务
	Event_UpDataOldService Event_Key = "UpDataOldService" //发现新的服务
	Event_LoseService      Event_Key = "LoseService"      //丢失服务
	Event_RegistryStart    Event_Key = "RegistryStart"    //注册表系统启动成功
)

const (
	S_Category_SystemService   S_Category = "SystemService"   //系统服务类型
	S_Category_GateService     S_Category = "GateService"     //网关服务类型
	S_Category_BusinessService S_Category = "BusinessService" //业务服务器
)

//服务配置数据接口定义
type ServiceSttings struct {
	Id       string                            //服务Id
	Type     string                            //服务类型 (相同的服务可以启动多个)
	Tag      string                            //服务集群标签 (相同标签的集群服务可以互相发现和发现)
	Category S_Category                        //服务列表 (用于区分集群服务下相似业务功能的服务器 例如:游戏服务器)
	Ip       string                            //服务所在Ip				()
	Port     int                               //服务rpcx监听端口
	Comps    map[string]map[string]interface{} //服务组件配置
	Sys      map[string]map[string]interface{} //服务系统配置
	Modules  map[string]map[string]interface{} //服务模块配置
}

//基础服务的接口设计
type IService interface {
	GetId() string                                              //获取服务id
	GetType() string                                            //获取服务类型
	GetVersion() string                                         //获取服务版本
	GetIp() string                                              //获取服务器运ip
	GetPort() int                                               //服务默认端口
	GetSettings() ServiceSttings                                //获取服务配置表信息
	Init(service IService) (err error)                          //初始化接口
	InitSys()                                                   //初始化系统
	OnInstallComp(cops ...IServiceComp)                         //组装服务组件
	Start() (err error)                                         //启动服务
	Run(mods ...IModule)                                        //运行服务
	Close(closemsg string)                                      //关闭服务
	Destroy() (err error)                                       //销毁服务
	GetComp(CompName S_Comps) (comp IServiceComp, err error)    //获取组件
	GetModule(ModuleName M_Modules) (module IModule, err error) //获取模块
}

//基础服务组件的接口设计
type IServiceComp interface {
	GetName() S_Comps
	NewOptions() (options ICompOptions)
	Init(service IService, comp IServiceComp, options ICompOptions) (err error)
	Start() (err error)
	Destroy() (err error)
}

//基础模块的接口设计
type IModule interface {
	GetType() M_Modules                                                        //模块类型
	NewOptions() (options IModuleOptions)                                      //返回模块的参数类型 用于自动序列化模块参数对象
	Init(service IService, module IModule, options IModuleOptions) (err error) //服务运行时 优先执行完所有模块的Init 函数
	OnInstallComp()                                                            //模块启动时在Init中有限执行模块的组件转载函数 初始化组件列表
	Start() (err error)                                                        //服务初始化所有模块后在同一执行各个模块的启动函数
	Run(closeSig chan bool) (err error)                                        //模块需要独立携程运行时可以重构此接口
	Destroy() (err error)                                                      //模块销毁接口
}

//服务组件参数对象接口
type ICompOptions interface {
	LoadConfig(settings map[string]interface{}) (err error) //序列化参数接口
}

//模块参数对象接口
type IModuleOptions interface {
	LoadConfig(settings map[string]interface{}) (err error) //序列化参数接口
}

//模块组件接口设计
type IModuleComp interface {
	Init(service IService, module IModule, comp IModuleComp, options IModuleOptions) (err error) //组件初始化函数
	Start() (err error)                                                                          //组件启动函数
	Destroy() (err error)                                                                        //组件销毁函数
}

//用户基础会话对象设计接口
type IUserSession interface {
	GetSessionId() string                                             //用户会话连接id 标识唯一的socket对象
	GetIP() string                                                    //用户的ip信息
	GetGateId() string                                                //用户所在的网关服务地址
	SendMsg(comdId uint16, msgId uint16, msg interface{}) (err error) //向用户发送消息
	Close() (err error)                                               //断开用户的连接
}

//服务后台运维接口 动态修改服务或模块配置参数
type IServiceMonitor interface {
	IModule
	RegisterServiceSettingItem(comp S_Comps, name string, iswrite bool, value interface{}, f func(newvalue string) (err error))    //注册服务级别的Setting
	RegisterModuleSettingItem(module M_Modules, name string, iswrite bool, value interface{}, f func(newvalue string) (err error)) //注册模块级别的Setting
}

//服务监控 Monitor 数据
type (
	SettingItem struct {
		ItemName string
		IsWrite  bool
		Data     interface{}
	}
	ServiceMonitor struct { //服务监听
		ServiceId       string                       //服务Id
		ServiceType     string                       //服务类型
		ServiceCategory S_Category                   //服务列表
		ServiceVersion  string                       //服务版本
		ServiceTag      string                       //服务集群
		Pid             int32                        //进程Id
		Pname           string                       //进程名称
		MemoryUsed      float64                      //内存使用量
		CpuUsed         float64                      //Cpu使用量
		TotalGoroutine  int                          //总的协程数
		CurrPreWeight   float64                      //服务权重
		CompsSetting    map[S_Comps]*CompsSetting    //服务器配置信息
		SysSetting      map[string]*SysSetting       //服务器系统配置信息
		ModuleMonitor   map[M_Modules]*ModuleMonitor //模块监听信息
	}
	CompsSetting struct { //模块监听
		CompName string                  //系统名称
		Setting  map[string]*SettingItem //系统配置信息
	}
	SysSetting struct { //模块监听
		SysName string                  //系统名称
		Setting map[string]*SettingItem //系统配置信息
	}
	ModuleMonitor struct { //模块监听
		ModuleName M_Modules               //模块名称
		Setting    map[string]*SettingItem //模块配置信息
	}
)
